/* cdDialog.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "cdDialog.h"
#include "artist.h"
#include "artistEntry.h"
#include "database.h"
#include "dataCell.h"
#include "cdReader.h"
#include "songDialog.h"
#include "cd.h"
#include "cdRecordDialog.h"
#include "formDef.h"
#include "defaultForms.h"
#include "widgetSize.h"
#include "format.h"
#include "fix_capitalization.h"
#include "string_utils.h"
#include "extract_words.h"
#include "playlistDialog.h"

#include <ctype.h>

CDTrackData::
CDTrackData() :
  _songs(TI_songs),
  _songs_extension(TI_songs_extension),
  _load(false)
{
}

void CDTrackData::
interpret_name(CDInterpretNameType type) {
  switch (type) {
  case IN_title:
    _songs[SFI_name] = DataCell(_orig_name);
    _songs[SFI_performer] = DataCell(_orig_performer);
    _songs[SFI_writer] = DataCell(_orig_writer);
    return;

  case IN_performer_slash_title:
    extract_artist(true, true, "/");
    return;

  case IN_writer_slash_title:
    extract_artist(false, true, "/");
    return;

  case IN_performer_hyphen_title:
    extract_artist(true, true, "-");
    return;

  case IN_writer_hyphen_title:
    extract_artist(false, true, "-");
    return;

  case IN_performer_dash_title:
    extract_artist(true, true, "--");
    return;

  case IN_writer_dash_title:
    extract_artist(false, true, "--");
    return;

  case IN_performer_colon_title:
    extract_artist(true, true, ":");
    return;

  case IN_writer_colon_title:
    extract_artist(false, true, ":");
    return;

  case IN_performer_spaces_title:
    extract_artist(true, true, "  ");
    return;

  case IN_writer_spaces_title:
    extract_artist(false, true, "  ");
    return;

  case IN_performer_TITLE:
    extract_artist_caps(true, true);
    return;

  case IN_writer_TITLE:
    extract_artist_caps(false, true);
    return;

  case IN_title_slash_performer:
    extract_artist(true, false, "/");
    return;

  case IN_title_slash_writer:
    extract_artist(false, false, "/");
    return;

  case IN_title_hyphen_performer:
    extract_artist(true, false, "-");
    return;

  case IN_title_hyphen_writer:
    extract_artist(false, false, "-");
    return;

  case IN_title_dash_performer:
    extract_artist(true, false, "--");
    return;

  case IN_title_dash_writer:
    extract_artist(false, false, "--");
    return;

  case IN_title_colon_performer:
    extract_artist(true, false, ":");
    return;

  case IN_title_colon_writer:
    extract_artist(false, false, ":");
    return;

  case IN_title_spaces_performer:
    extract_artist(true, false, "  ");
    return;

  case IN_title_spaces_writer:
    extract_artist(false, false, "  ");
    return;

  case IN_title_paren_performer:
    extract_artist(true, false, "(", ")");
    return;

  case IN_title_paren_writer:
    extract_artist(false, false, "(", ")");
    return;

  case IN_title_bracket_performer:
    extract_artist(true, false, "[", "]");
    return;

  case IN_title_bracket_writer:
    extract_artist(false, false, "[", "]");
    return;
  }

  cerr << "Unexpected InterpretNameType: " << type << "\n";
}

// Extract the title and artist from the CDDB title name, based on
// explicit punctuation between the two.
void CDTrackData::
extract_artist(bool performer, bool artist_first, const string &separator,
	       const string &trail) {
  int sep = _orig_name.find(separator);
  int second_part = string::npos;
  int sep_len = string::npos;

  if (sep != string::npos) {
    second_part = sep + separator.length();
    int trp = string::npos;
    if (!trail.empty()) {
      trp = _orig_name.find(trail, sep);
      if (trp == string::npos) {
	// If we didn't find the trailing character, the leading
	// character doesn't count.
	sep = string::npos;
      } else {
	sep_len = trp - second_part;
      }
    }
  }

  if (sep != string::npos) {
    string first = trim_blanks(_orig_name.substr(second_part, sep_len));
    string second = trim_blanks(_orig_name.substr(0, sep));

    string new_name;
    TableDataProxy artist;

    if (artist_first) {
      new_name = first;
      artist = find_artist(second);
    } else {
      artist = find_artist(first);
      new_name = second;
    }

    _songs[SFI_name] = DataCell(new_name);
    if (performer) {
      _songs[SFI_performer] = DataCell(artist);
      _songs[SFI_writer] = DataCell(_orig_writer);
    } else {
      _songs[SFI_performer] = DataCell(_orig_performer);
      _songs[SFI_writer] = DataCell(artist);
    }
  }
}

// A utility function for extract_artist_caps(), below.  If the string
// is ALL CAPITALS, this returns the number of letters that are
// capital.  If there are some lowercase letters, this returns 0.
static int
count_all_caps(const string &word) {
  int count_caps = 0;
  for (string::const_iterator si = word.begin(); si != word.end(); ++si) {
    if (islower(*si)) {
      return 0;
    } else if (isupper(*si)) {
      count_caps++;
    }
  }

  return count_caps;
}

// Extract the title and artist from the CDDB title name, where there
// is no explicit punctuation between the two but or one the other is
// written in ALL CAPS.  Sheesh.
void CDTrackData::
extract_artist_caps(bool performer, bool title_caps) {
  vector<string> words = extract_words(_orig_name);

  string lower_string;
  string upper_string;

  int start_caps = -1;

  // Search for the beginning of the all-capitalized words.
  int maybe_caps = -1;
  for (int i = 0; i < words.size() && start_caps < 0; i++) {
    int num_caps = count_all_caps(words[i]);
    if (num_caps == 1) {
      // If we only have one capital letter, but no lowercase letters,
      // that might be an all-capitalized word.  Mark it for later.
      if (maybe_caps < 0) {
	maybe_caps = i;
      }

    } else if (num_caps > 1) {
      start_caps = (maybe_caps > 0) ? maybe_caps : i;

    } else {
      maybe_caps = -1;
    }
  }

  if (start_caps < 0) {
    return;
  }

  if (start_caps == 0) {
    int end_caps = words.size();

    // The first word was all-capitalized.  Now search for the end.
    int i;
    for (i = 0; i < words.size() && end_caps == words.size(); i++) {
      if (count_all_caps(words[i]) == 0) {
	end_caps = i;
      }
    }

    for (i = 0; i < end_caps; i++) {
      upper_string += " " + words[i];
    }
    for (i = end_caps; i < words.size(); i++) {
      lower_string += " " + words[i];
    }

  } else {

    int i;
    for (i = 0; i < start_caps; i++) {
      lower_string += " " + words[i];
    }
    for (i = start_caps; i < words.size(); i++) {
      upper_string += " " + words[i];
    }
  }

  upper_string = ::fix_capitalization(trim_blanks(upper_string));
  lower_string = trim_blanks(lower_string);

  if (title_caps) {
    _songs[SFI_name] = DataCell(upper_string);
    if (performer) {
      _songs[SFI_performer] = DataCell(find_artist(lower_string));
      _songs[SFI_writer] = DataCell(_orig_writer);
    } else {
      _songs[SFI_performer] = DataCell(_orig_performer);
      _songs[SFI_writer] = DataCell(find_artist(lower_string));
    }

  } else {
    _songs[SFI_name] = DataCell(lower_string);
    if (performer) {
      _songs[SFI_performer] = DataCell(find_artist(upper_string));
      _songs[SFI_writer] = DataCell(_orig_writer);
    } else {
      _songs[SFI_performer] = DataCell(_orig_performer);
      _songs[SFI_writer] = DataCell(find_artist(upper_string));
    }
  }    
}

void CDTrackData::
fix_capitalization() {
  _songs[SFI_name] = 
    DataCell(::fix_capitalization(_songs[SFI_name].get_string()));

  if (!_songs[SFI_performer].is_null()) {
    const TableDataProxy &artist = _songs[SFI_performer].get_foreign(TI_artists);
    if (artist.is_new_data()) {
      // We can only fiddle with the capitalization of an artist name
      // if it's a new record.
      _songs[SFI_performer] = DataCell(find_artist(::fix_capitalization(get_formatted_artist_name(artist))));
    }
  }

  if (!_songs[SFI_writer].is_null()) {
    const TableDataProxy &artist = _songs[SFI_writer].get_foreign(TI_artists);
    if (artist.is_new_data()) {
      // We can only fiddle with the capitalization of an artist name
      // if it's a new record.
      _songs[SFI_writer] = DataCell(find_artist(::fix_capitalization(get_formatted_artist_name(artist))));
    }
  }
}

void CDTrackData::
set_orig_values() {
  _orig_name = _songs[SFI_name].get_string();
  _orig_performer = _songs[SFI_performer].get_foreign(TI_artists);
  _orig_writer = _songs[SFI_writer].get_foreign(TI_artists);
}

CDTrackList::
CDTrackList() : Playlist(TI_songs) {
  _cd_dialog = NULL;
  _columns.push_back(Column("Track", DataCell::DC_string));
  _columns.push_back(Column("Performer", DataCell::DC_foreign_key, TI_artists));
  _columns.push_back(Column("Writer", DataCell::DC_foreign_key, TI_artists));
  _columns.push_back(Column("Name", DataCell::DC_string));
  _columns.push_back(Column("Length", DataCell::DC_elapsed_time));
  _columns.push_back(Column("Load?", DataCell::DC_string));
}

TableDataProxy CDTrackList::
get_table_data(int row) {
  assert(row >= 0 && row < _tracks.size());
  return _tracks[row]._songs;
}

// This gets called just before popping up the indicated SongDialog.
// It gives us a chance to store the song data appropriate for it,
// instead of the default behavior of looking this data up in the
// database.
void CDTrackList::
fill_dialog_data(int row, Dataset &data) {
  data[TI_songs] = _tracks[row]._songs;
  data[TI_songs_extension] = _tracks[row]._songs_extension;
}

// This gets called just after popping up the indicated SongDialog.
// It gives us a chance to set any fields on it.
void CDTrackList::
dialog_created(int row, PalomaDialog *dialog) {
  assert(row >= 0 && row < _tracks.size());
  SongDialog *song_dialog = (SongDialog *)dialog;
  song_dialog->set_load(_tracks[row]._load);
}

void CDTrackList::
dialog_accepted(int row, PalomaDialog *dialog) {
  assert(row >= 0 && row < _tracks.size());
  SongDialog *song_dialog = (SongDialog *)dialog;
  const Dataset &dialog_data = song_dialog->get_data();

  _tracks[row]._songs = dialog_data[TI_songs];
  _tracks[row]._songs_extension = dialog_data[TI_songs_extension];
  _tracks[row]._load = song_dialog->get_load();
  _tracks[row].set_orig_values();

  update_clist(row);
}

void CDTrackList::
update_row(int row) {
  assert(row >= 0 && row < _tracks.size());
  const TableDataProxy &song = _tracks[row]._songs;

  string load;
  if (_tracks[row]._load) {
    load = "Yes";
  } else {
    if (!song[SFI_loaded_time].is_null()) {
      // The song is already loaded.
      load = "--";
    } else {
      load = "No";
    }
  }

  ColumnValues &cv = _rows[row]._column_values;
  cv.clear();
  cv.push_back(DataCell(row + 1));
  cv.push_back(song[SFI_performer]);
  cv.push_back(song[SFI_writer]);
  cv.push_back(song[SFI_name]);
  cv.push_back(song[SFI_length]);
  cv.push_back(DataCell(load));

  if (_cd_dialog != NULL) {
    _cd_dialog->check_load_all();
  }
}

void CDTrackList::
update_clist(int row) {
  update_row(row);
  Playlist::update_clist(row);
}

CDTrackListProxy::
CDTrackListProxy(CDTrackList *ptr) : PlaylistProxy(ptr) {
}

CDTrackListProxy::
CDTrackListProxy(const CDTrackListProxy &copy) : PlaylistProxy(copy) {
}

CDTrackListProxy &CDTrackListProxy::
operator = (const CDTrackListProxy &copy) {
  PlaylistProxy::operator = (copy);
  return *this;
}

CDTrackData &CDTrackListProxy::
get_track(int row) {
  return ((CDTrackList *)get_ptr())->_tracks[row];
}

void CDTrackListProxy::
set_cd_dialog(CDDialog *dialog) {
  ((CDTrackList *)get_ptr())->_cd_dialog = dialog;
}

CDDialog::
CDDialog(const Dataset &data, int form_id, 
	 const CDTrackListProxy &track_list, CDReader *cd_reader) :
  PalomaDialog(data), _form_id(form_id),
  _track_list(track_list), _cd_reader(cd_reader),
  _menu("<cd>"),
  _cd_title_label("CD Title:"),
  _performer_label("Performer:"),
  _performer_entry(data),
  _writer_label("Writer / Composer:"),
  _writer_entry(data),
  _below_clist(2, 3),
  _capitalization_button("Fix Capitalization"),
  _show_performer_clist_check("Show Performer"),
  _show_writer_clist_check("Show Writer"),
  _load_all_label("Load all"),
  _load_none("Load none"),
  _controls(data),
  _ok_button("OK"),
  _cancel_button("Cancel")
{
  DefaultForms::check_default(_form_id, TI_cds);
  
  _track_list.set_cd_dialog(this);
  _cd_title_entry.set_text(_data[TI_cds][CFI_title].get_string());
  _performer_entry.set_artist(_data[TI_cds][CFI_performer].get_foreign(TI_artists));
  _writer_entry.set_artist(_data[TI_cds][CFI_writer].get_foreign(TI_artists));

  _interpret_name_pick.add_choice(IN_title, "Title");
  _interpret_name_pick.add_choice(IN_performer_slash_title, "Performer / Title");
  _interpret_name_pick.add_choice(IN_writer_slash_title, "Writer / Title");
  _interpret_name_pick.add_choice(IN_performer_hyphen_title, "Performer - Title");
  _interpret_name_pick.add_choice(IN_writer_hyphen_title, "Writer - Title");
  _interpret_name_pick.add_choice(IN_performer_dash_title, "Performer -- Title");
  _interpret_name_pick.add_choice(IN_writer_dash_title, "Writer -- Title");
  _interpret_name_pick.add_choice(IN_performer_colon_title, "Performer : Title");
  _interpret_name_pick.add_choice(IN_writer_colon_title, "Writer : Title");
  _interpret_name_pick.add_choice(IN_performer_spaces_title, "Performer   Title");
  _interpret_name_pick.add_choice(IN_writer_spaces_title, "Writer   Title");
  _interpret_name_pick.add_choice(IN_performer_TITLE, "Performer TITLE");
  _interpret_name_pick.add_choice(IN_writer_TITLE, "Writer TITLE");
  _interpret_name_pick.add_choice(IN_title_slash_performer, "Title / Performer");
  _interpret_name_pick.add_choice(IN_title_slash_writer, "Title / Writer");
  _interpret_name_pick.add_choice(IN_title_hyphen_performer, "Title - Performer");
  _interpret_name_pick.add_choice(IN_title_hyphen_writer, "Title - Writer");
  _interpret_name_pick.add_choice(IN_title_dash_performer, "Title -- Performer");
  _interpret_name_pick.add_choice(IN_title_dash_writer, "Title -- Writer");
  _interpret_name_pick.add_choice(IN_title_colon_performer, "Title : Performer");
  _interpret_name_pick.add_choice(IN_title_colon_writer, "Title : Writer");
  _interpret_name_pick.add_choice(IN_title_spaces_performer, "Title   Performer");
  _interpret_name_pick.add_choice(IN_title_spaces_writer, "Title   Writer");
  _interpret_name_pick.add_choice(IN_title_paren_performer, "Title (Performer)");
  _interpret_name_pick.add_choice(IN_title_paren_writer, "Title (Writer)");
  _interpret_name_pick.add_choice(IN_title_bracket_performer, "Title [Performer]");
  _interpret_name_pick.add_choice(IN_title_bracket_writer, "Title [Writer]");
  _interpret_name_pick.finish_choices();

  if (_form_id >= 0) {
    _form.layout_form(_form_id, _data);
    _form.show();
  }

  _clist = _track_list.make_clist();
  _track_list.set_next_form_id(get_related_form_id(_form_id, TI_songs));
  _track_list.set_cd_reader(_cd_reader);

  _reload_all = false;
  check_load_all();

  layout_window();
  setup();
}

CDDialog::
~CDDialog() {
  if (_cd_reader != NULL) {
    // We must call clear_player() before destructing, since the
    // PlaylistControls object has a pointer to our cd_reader and may
    // try to access it when it destructs.
    _controls.clear_player();
  }
}

// Reads a CD physically in the drive (as defined by the CDReader) and
// returns a CDDialog object that reflects that CD.  Note that the
// CDDialog becomes the owner of the CDReader pointer.  It will delete
// it on destruction.

CDDialog *CDDialog::
make_dialog(const Dataset &orig_data, CDReader *cd_reader) {
#ifndef HAVE_CDAUDIO
  cerr << "No CDReader available!\n";
  abort();
  return NULL;
#else   // HAVE_CDAUDIO

  Dataset data = orig_data;
  CDTrackList *track_list = new CDTrackList;

  string cddb_id = cd_reader->get_cddb_id();
  int num_tracks = cd_reader->get_num_tracks();

  data[TI_cds].clear();
  data[TI_cds_extension].clear();

  // Does this CD already exist in the database?
  int cd_id = -1;
  if (!cddb_id.empty()) {
    QueryResult qr = music_db->sql_command
      ("SELECT cd_id, cddb, performer, writer, title "
       "FROM cds WHERE cddb=" + DataCell(cddb_id));

    if (qr.get_num_rows() != 0) {
      // There is already at least one CD in the database with the
      // matching CDDB id.  Odds are pretty good that's our match.
      // However, it's possible it's not.  Scan through each of the
      // found matches until we find the one that is definitely the
      // right CD.
      int ci;
      int num_rows = qr.get_num_rows();
      for (ci = 0; ci < num_rows && cd_id == -1; ci++) {
	int id = qr.get_cell(ci, 0).get_int();

	QueryResult qr2 = music_db->sql_command
	  ("SELECT cd_tracks.song_id,"
	   " songs.name, songs.length, songs.performer, songs.writer,"
	   " songs.loaded_time, songs.file_type, songs.volume_id,"
	   " songs.filename"
	   " FROM cd_tracks, songs"
	   " WHERE cd_tracks.cd_id=" + DataCell(id) +
	   " AND cd_tracks.song_id=songs.song_id"
	   " ORDER BY cd_tracks.track_num");
	
	// We should probably also verify length of tracks here.
	
	if (qr2.get_num_rows() == num_tracks) {
	  // The CDDB id and number of tracks from the database match
	  // the CD in the drive; it must be the right CD.  Retrieve all
	  // the information from the database.

	  cd_id = id;
	  data[TI_cds].set_primary_key(cd_id, false);
	  data[TI_cds].set_db_value(CFI_cddb, qr.get_cell(ci, 1));
	  data[TI_cds].set_db_value(CFI_performer, qr.get_cell(ci, 2));
	  data[TI_cds].set_db_value(CFI_writer, qr.get_cell(ci, 3));
	  data[TI_cds].set_db_value(CFI_title, qr.get_cell(ci, 4));

	  data[TI_cds_extension] = TableDataProxy(TI_cds_extension, cd_id);
	  
	  // Now fill in all the tracks.
	  for (int t = 0; t < num_tracks; t++) {
	    int song_id = qr2.get_cell(t, 0).get_int();
	    CDTrackData track;
	    track._songs.set_primary_key(song_id, false);
	    track._songs.set_db_value(SFI_name, qr2.get_cell(t, 1));
	    track._songs.set_db_value(SFI_length, qr2.get_cell(t, 2));
	    track._songs.set_db_value(SFI_performer, qr2.get_cell(t, 3));
	    track._songs.set_db_value(SFI_writer, qr2.get_cell(t, 4));
 	    track._songs.set_db_value(SFI_loaded_time, qr2.get_cell(t, 5));
 	    track._songs.set_db_value(SFI_file_type, qr2.get_cell(t, 6));
 	    track._songs.set_db_value(SFI_volume_id, qr2.get_cell(t, 7));
 	    track._songs.set_db_value(SFI_filename, qr2.get_cell(t, 8));
	    track._songs_extension = TableDataProxy(TI_songs_extension, song_id);
	    track.set_orig_values();

	    track_list->_tracks.push_back(track);
	    track_list->_rows.push_back(CDTrackList::Row(song_id));
	    track_list->update_row(t);
	  }
	}
      }
    }
  }

  if (cd_id == -1) {
    // This CD was not previously in the database; fill in all the
    // data now.
    cd_reader->query_cddb();
    string title = cd_reader->get_title();

    data[TI_cds][CFI_cddb] = DataCell(cddb_id);
    data[TI_cds][CFI_title] = DataCell(title, !title.empty());
    data[TI_cds][CFI_performer] = DataCell(find_artist(cd_reader->get_artist()));

    for (int t = 0; t < num_tracks; t++) {
      CDTrackData track;
      string name = cd_reader->get_track_name(t);
      string artist = cd_reader->get_track_artist(t);
      track._songs[SFI_name] = DataCell(name, !name.empty());
      track._songs[SFI_length] = DataCell(cd_reader->get_track_length(t));

      if (artist.empty()) {
	track._songs[SFI_performer] = data[TI_cds][CFI_performer];
      } else {
	track._songs[SFI_performer] = DataCell(find_artist(artist));
      }
      track.set_orig_values();
      track_list->_tracks.push_back(track);
      track_list->_rows.push_back(CDTrackList::Row(-1));
      track_list->update_row(t);
    }
  }  

  return new CDDialog(data, -1, track_list, cd_reader);
#endif  // HAVE_CDAUDIO
}

// Creates a dialog for the CD indicated in the current TI_cd entry in
// the dataset.
CDDialog *CDDialog::
make_dialog(const Dataset &data, int form_id) {
  CDTrackList *track_list = new CDTrackList;

  int id = data[TI_cds].get_primary_key();

  QueryResult qr2 = music_db->sql_command
    ("SELECT cd_tracks.song_id, "
     "songs.name, songs.length, songs.performer, songs.writer,"
     " songs.loaded_time, songs.file_type, songs.volume_id,"
     " songs.filename"
     " FROM cd_tracks, songs"
     " WHERE cd_tracks.cd_id=" + DataCell(id) +
     " AND cd_tracks.song_id=songs.song_id"
     " ORDER BY track_num");
	
  int num_tracks = qr2.get_num_rows();
	  
  for (int t = 0; t < num_tracks; t++) {
    int song_id = qr2.get_cell(t, 0).get_int();
    CDTrackData track;
    track._songs.set_primary_key(song_id, false);
    track._songs.set_db_value(SFI_name, qr2.get_cell(t, 1));
    track._songs.set_db_value(SFI_length, qr2.get_cell(t, 2));
    track._songs.set_db_value(SFI_performer, qr2.get_cell(t, 3));
    track._songs.set_db_value(SFI_writer, qr2.get_cell(t, 4));
    track._songs.set_db_value(SFI_loaded_time, qr2.get_cell(t, 5));
    track._songs.set_db_value(SFI_file_type, qr2.get_cell(t, 6));
    track._songs.set_db_value(SFI_volume_id, qr2.get_cell(t, 7));
    track._songs.set_db_value(SFI_filename, qr2.get_cell(t, 8));
    track._songs_extension = TableDataProxy(TI_songs_extension, song_id);
    track.set_orig_values();
    track_list->_tracks.push_back(track);
    track_list->_rows.push_back(CDTrackList::Row(song_id));
    track_list->update_row(t);
  }

  return new CDDialog(data, form_id, track_list, NULL);
}

void CDDialog::
record_db() {
  // First, record the CD header information and extensions.
  bool new_data = _data[TI_cds].is_new_data();

  int cd_id = _data[TI_cds].record_db();
  _data[TI_cds_extension].record_db(cd_id);  

  // Then, we need to record all of the tracks.
  int t;
  for (t = 0; t < _track_list.get_num_rows(); t++) {
    int song_id = _track_list.get_track(t)._songs.record_db();
    _track_list.get_track(t)._songs_extension.record_db(song_id);

    if (new_data) {
      // If this CD wasn't in the database before, we need to add a
      // record that links in each track.
      music_db->sql_command
	("INSERT INTO cd_tracks (cd_id, track_num, song_id) "
	 " VALUES (" + 
	 DataCell(cd_id) + ", " +
	 DataCell(t+1) + ", " +
	 DataCell(song_id) + ")");
    }
  }
}

void CDDialog::
check_load_all() {
  // If all of the tracks have been loaded or marked as loaded except
  // at least one track which has been loaded previously, then the
  // "Load all" button should become a "Reload all" button.
  bool any_previous = false;
  bool all_loaded = true;

  int num_rows = _track_list.get_num_rows();
  for (int r = 0; r < num_rows; r++) {
    const CDTrackData &track = _track_list.get_track(r);

    bool loaded = track._load;
    if (!track._songs[SFI_loaded_time].is_null()) {
      loaded = true;
      any_previous = true;
    }
    all_loaded = all_loaded && loaded;
  }

  if (all_loaded && any_previous) {
    if (!_reload_all) {
      _reload_all = true;
      _load_all_label.set_text("Reload all");
    }
  } else {
    if (_reload_all) {
      _reload_all = false;
      _load_all_label.set_text("Load all");
    }
  }
}

void CDDialog::
menu_same_performer() {
  PlaylistDialog::make_dialog
    (_data, -1, 
     "SELECT cds.cd_id, cds.title"
     " FROM cds"
     " WHERE cds.performer=" + DataCell(_performer_entry.get_artist()));
}  

void CDDialog::
menu_make_playlist() {
  PlaylistDialog::make_dialog
    (_data, -1, 
     "SELECT songs.song_id, songs.performer, songs.name"
     " FROM songs, cd_tracks"
     " WHERE cd_tracks.song_id=songs.song_id"
     " AND cd_tracks.cd_id=" + DataCell(_data[TI_cds].get_primary_key()));
}  

void CDDialog::
interpret_name_changed() {
  // Reinterpret the titles of the tracks according to the selection
  // in the interpret_name combo box.

  if (!_interpret_name_pick.has_item()) {
    return;
  }

  CDInterpretNameType type = 
    (CDInterpretNameType)_interpret_name_pick.get_item();

  for (int t = 0; t < _track_list.get_num_rows(); t++) {
    _track_list.get_track(t).interpret_name(type);
    _track_list.update_clist(t);
  }
}

void CDDialog::
capitalization_button_clicked() {
  // Regularize the capitalization of the track names, CD title, and
  // associated artists.

  string title = _cd_title_entry.get_text();

  title = fix_capitalization(title);
  _data[TI_cds][CFI_title] = DataCell(title);
  _cd_title_entry.set_text(title);

  for (int t = 0; t < _track_list.get_num_rows(); t++) {
    _track_list.get_track(t).fix_capitalization();
    _track_list.update_clist(t);
  }
}

void CDDialog::
show_performer_clist_clicked() {
  // Toggle the showing of the performer in the track list window.
  if (_show_performer_clist_check.get_active()) {
    if (_show_writer_clist_check.get_active()) {
      // Hide the writer first.
      _show_writer_clist_check.set_active(false);
    }

    // Show the performer.
    _clist->set_column_width(3, 150);
    _clist->set_column_visibility(1, true);
  } else {
    // Hide the performer.
    _clist->set_column_width(3, 270);
    _clist->set_column_visibility(1, false);
  }
}

void CDDialog::
show_writer_clist_clicked() {
  // Toggle the showing of the writer in the track list window.
  if (_show_writer_clist_check.get_active()) {
    if (_show_performer_clist_check.get_active()) {
      // Hide the performer first.
      _show_performer_clist_check.set_active(false);
    }

    // Show the writer.
    _clist->set_column_width(3, 150);
    _clist->set_column_visibility(2, true);
  } else {
    // Hide the writer.
    _clist->set_column_width(3, 270);
    _clist->set_column_visibility(2, false);
  }
}

void CDDialog::
performer_activated() {
  // We just changed the performer on the CD; change each song to
  // match (if it matched previously).

  const TableDataProxy &old_performer = 
    _data[TI_cds][CFI_performer].get_foreign(TI_artists);
  const TableDataProxy &new_performer =
    _performer_entry.get_artist();

  string new_name = get_formatted_artist_name(new_performer);

  if (old_performer != new_performer) {
    for (int t = 0; t < _track_list.get_num_rows(); t++) {
      const TableDataProxy &song_performer = 
	_track_list.get_track(t)._songs[SFI_performer].get_foreign(TI_artists);

      if (!song_performer.has_table() ||
	  song_performer == old_performer) {
	_track_list.get_track(t)._songs[SFI_performer] = DataCell(new_performer);
	_clist->set_text(t, 1, new_name);
      }
    }

    _data[TI_cds][CFI_performer] = DataCell(new_performer);
  }
}

void CDDialog::
writer_activated() {
  // We just changed the writer on the CD; change each song to
  // match (if it matched previously).

  const TableDataProxy &old_writer = 
    _data[TI_cds][CFI_writer].get_foreign(TI_artists);
  const TableDataProxy &new_writer =
    _writer_entry.get_artist();

  string new_name = get_formatted_artist_name(new_writer);

  if (old_writer != new_writer) {
    for (int t = 0; t < _track_list.get_num_rows(); t++) {
      const TableDataProxy &song_writer = 
	_track_list.get_track(t)._songs[SFI_writer].get_foreign(TI_artists);

      if (!song_writer.has_table() ||
	  song_writer == old_writer) {
	_track_list.get_track(t)._songs[SFI_writer] = DataCell(new_writer);
	_clist->set_text(t, 2, new_name);
      }
    }

    _data[TI_cds][CFI_writer] = DataCell(new_writer);
  }
}

void CDDialog::
load_all_clicked() {
  for (int t = 0; t < _track_list.get_num_rows(); t++) {
    // Only flag the songs that aren't already loaded.
    if (_track_list.get_track(t)._songs[SFI_loaded_time].is_null() ||
	_reload_all) {
      _track_list.get_track(t)._load = true;
    }
    _track_list.update_clist(t);
  }
  check_load_all();
}

void CDDialog::
load_none_clicked() {
  for (int t = 0; t < _track_list.get_num_rows(); t++) {
    _track_list.get_track(t)._load = false;
    _track_list.update_clist(t);
  }
  check_load_all();
}

void CDDialog::
ok_button_clicked() {
  _data[TI_cds][CFI_title] = DataCell(_cd_title_entry.get_text());
  _data[TI_cds][CFI_performer] = DataCell(_performer_entry.get_artist());
  _data[TI_cds][CFI_writer] = DataCell(_writer_entry.get_artist());
  _form.update();
  record_db();

  accepted();
  load_tracks();
  delete_self();
}

void CDDialog::
cancel_button_clicked() {
  delete_self();
}

void CDDialog::
layout_window() {
  set_default_size(450, 0);
  set_title("Define CD");

  // The menu bar.
  _menubar = _menu.get_menubar_widget("");
  assert(_menubar.get_object() != NULL);

  _menu.create_item("/File/CD's by this performer", "", "",
		    ItemFactoryConnector<CDDialog, int>(this, &menu_same_performer));
  _menu.create_item("/File/Make a playlist", "", "",
		    ItemFactoryConnector<CDDialog, int>(this, &menu_make_playlist));
  _menu.create_item("/File/sep", "", "<Separator>", 0);
  _menu.create_item("/File/Save", "", "",
		    ItemFactoryConnector<CDDialog, int>(this, &record_db));
  _menu.create_item("/File/Close", "", "",
		    ItemFactoryConnector<CDDialog, int>(this, &ok_button_clicked));

  vbox_.pack_start(*_menubar.get_object(), false, false);
  _menubar.get_object()->show();

  // The box for the main set of widgets.
  vbox_.pack_start(_box1);
  _box1.show();
  _box1.set_border_width(8);

  // The vbox to hold most of the entries.
  _box1.pack_start(_box2, false, false, 0);
  _box2.show();

  // CD title, label and entry.
  _cd_title_label.set_alignment(0, 0);
  _box2.pack_start(_cd_title_label, false, false, 0);
  _box2.pack_start(_cd_title_entry, true, true, 0);
  _cd_title_label.show();
  _cd_title_entry.show();

  // Performer, label and entry.
  _performer_label.set_alignment(0, 0);
  _box2.pack_start(_performer_label, false, false, 0);
  _box2.pack_start(_performer_entry, true, true, 0);
  _performer_label.show();
  _performer_entry.show();
  connect_to_method(_performer_entry.activate, this, &performer_activated);

  // Writer, label and entry.
  _writer_label.set_alignment(0, 0);
  _box2.pack_start(_writer_label, false, false, 0);
  _box2.pack_start(_writer_entry, true, true, 0);
  _writer_label.show();
  _writer_entry.show();
  connect_to_method(_writer_entry.activate, this, &writer_activated);

  // The user's custom form, if defined.
  _box2.pack_start(_form, true, true, 0);

  // The list of tracks.
  request_initial_size(_track_window, 0, 150);
  _track_window.set_policy(GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  _box1.pack_start(_track_window, true, true, 0);
  _track_window.show();

  _track_window.add_with_viewport(*_clist);

  _clist->column_titles_passive();
  _clist->set_column_justification(0, GTK_JUSTIFY_RIGHT);
  _clist->set_column_justification(4, GTK_JUSTIFY_RIGHT);
  _clist->set_column_justification(5, GTK_JUSTIFY_CENTER);
  _clist->set_column_width(0, 30);
  _clist->set_column_width(1, 112);
  _clist->set_column_width(2, 112);
  _clist->set_column_width(3, 270);
  _clist->set_column_width(4, 40);
  _clist->set_column_width(5, 40);
  _clist->set_column_visibility(1, false);
  _clist->set_column_visibility(2, false);
  _clist->set_column_visibility(5, _cd_reader != NULL);
  _clist->show();

  // The group of things just below the track list window.
  _box1.pack_start(_below_clist, false, false, 0);
  _below_clist.set_row_spacings(4);
  _below_clist.set_col_spacings(4);
  _below_clist.show();

  // The combo box to choose the interpretation of the track name.
  _below_clist.attach(_interpret_name_pick, 0, 1, 0, 1);
  _interpret_name_pick.show();
  connect_to_method(_interpret_name_pick.changed_item, this, 
		    &interpret_name_changed);

  // The button to fix capitalization errors.
  _below_clist.attach(_capitalization_button, 0, 1, 1, 2, 0, 0);
  _capitalization_button.show();
  connect_to_method(_capitalization_button.clicked, this, 
		    &capitalization_button_clicked);

  // The button to toggle showing the performer in the track list.
  _below_clist.attach(_show_performer_clist_check, 1, 2, 0, 1);
  _show_performer_clist_check.show();
  connect_to_method(_show_performer_clist_check.clicked, this, 
		    &show_performer_clist_clicked);

  // The button to toggle showing the writer in the track list.
  _below_clist.attach(_show_writer_clist_check, 1, 2, 1, 2);
  _show_writer_clist_check.show();
  connect_to_method(_show_writer_clist_check.clicked, this, 
		    &show_writer_clist_clicked);

  if (_cd_reader != NULL) {
    // The "Load all" button.
    _below_clist.attach(_load_all, 2, 3, 0, 1);
    _load_all.show();
    _load_all.add(_load_all_label);
    _load_all_label.show();
    connect_to_method(_load_all.clicked, this, &load_all_clicked);

    // The "Load none" button.
    _below_clist.attach(_load_none, 2, 3, 1, 2);
    _load_none.show();
    connect_to_method(_load_none.clicked, this, &load_none_clicked);
  }

  // The sound controls.
  _controls.layout(_track_list);
  _box1.pack_start(_controls, false, false, 0);
  _controls.show();

  // The OK button.
  _ok_button.set_usize(80, 36);
  action_area_.pack_start(_ok_button, true, false, 0);
  _ok_button.set_flags(GTK_CAN_DEFAULT);
  _ok_button.grab_default();
  _ok_button.show();
  connect_to_method(_ok_button.clicked, this, &ok_button_clicked);
  
  // The Cancel button.
  _cancel_button.set_usize(80, 36);
  action_area_.pack_start(_cancel_button, true, false, 0);
  _cancel_button.set_flags(GTK_CAN_DEFAULT);
  _cancel_button.show();
  connect_to_method(_cancel_button.clicked, this, &cancel_button_clicked);
}

void CDDialog::
load_tracks() {
#ifdef HAVE_CDAUDIO
  if (_cd_reader != NULL) {
    bool any_load = false;
    for (int t = 0; t < _track_list.get_num_rows() && !any_load; t++) {
      any_load = _track_list.get_track(t)._load;
    }
    
    if (any_load) {
      CDRecordDialog *dialog = new CDRecordDialog(_data, _track_list);
      dialog->begin();
    }
  }
#endif // HAVE_CDAUDIO
}
