/* cdRecordDialog.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#ifdef HAVE_CDAUDIO

#include "cdRecordDialog.h"
#include "cd.h"
#include "cdReader.h"
#include "song.h"
#include "infoMessage.h"
#include "database.h"
#include "volume.h"
#include "queryResult.h"
#include "format.h"

#include <unistd.h>
#include <time.h>

CDRecordDialog::
CDRecordDialog(const Dataset &data, const CDTrackListProxy &track_list) :
  PalomaDialog(data),
  _track_list(track_list),
  _cancel_button("Cancel")
{
  _cd_reader = _track_list.get_cd_reader();
  assert(_cd_reader != NULL);

  _cd_title.set_text(get_formatted_cd_name(_data[TI_cds]));

  // Count up the total time for all tracks to be read.
  _total_length = 0.0;
  int first_track = -1;
  for (int t = 0; t < _track_list.get_num_rows(); t++) {
    if (_track_list.get_track(t)._load) {
      // Here's a track to load.
      if (first_track < 0) {
	first_track = t;
      }

      _total_length += _track_list.get_track(t)._songs[SFI_length].get_float();
    }
  }

  // We should have found at least one track.
  assert(first_track >= 0);

  _current_length = 0.0;
  _current_track_index = first_track;
  _progress_bar.configure(0.0, 0.0, _total_length);
  _progress_bar.set_show_text(true);

  layout_window();

  // We don't call setup() yet, because we might want to blow away the
  // window before it ever comes up.  We'll call setup() below.

  _canceled = false;
}

void CDRecordDialog::
begin() {
  _volume = Volume(_data, _data[TI_configure][CFI_cdrom_load_volume].get_foreign(TI_volumes));
  _load_type = _data[TI_configure][CFI_cdrom_load_type].get_int();

  if (!_volume.is_valid()) {
    new InfoMessage("No disk volume defined for loading from CD-Rom.");
    delete_self();
    return;

  } else if (!_volume.exists()) {
    new InfoMessage("Volume " + _volume.get_description() +
		    " does not exist!");
    delete_self();
    return;

  } else if (!_volume.can_store_files()) {
    new InfoMessage("Cannot write to volume " +
		    _volume.get_description() + ".");
    delete_self();
    return;
  }

  if (_load_type < 0) {
    new InfoMessage("No load type defined for loading from CD-Rom.");
    delete_self();
    return;
  }

  QueryResult qr = music_db->sql_command
    ("SELECT extension FROM soundfile_exts WHERE soundfile_id=" +
     DataCell(_load_type) +
     " ORDER BY extension");
  if (qr.get_num_rows() > 0) {
    _extension = qr.get_cell(0, 0).get_string();
  }

  connect_to_method(_cd_reader->done_reading, this, &reader_done);

  TableDataProxy &song = _track_list.get_track(_current_track_index)._songs;

  _progress_bar.set_value(_current_length);
  _track_name.set_text(get_formatted_song_name(song));

  if (!start_reading()) {
    delete_self();
    return;
  }
  setup();
}

CDRecordDialog::
~CDRecordDialog() {
  assert(_cd_reader != NULL);
  // Eject the CD when we're done reading it.
  _cd_reader->eject();
}

void CDRecordDialog::
reader_done(bool success) {
  if (success) {
    if (access(_filename.c_str(), F_OK) != 0) {
      new InfoMessage("CD-reading process did not generate an output file.");
      success = false;
    }
  }

  if (success) {
    _filename = _volume.close_write_file(_filename, _basename);
    if (_filename.empty()) {
      new InfoMessage("Error trying to move file into bucket.");
      success = false;
    }
  }

  if (success) {
    TableDataProxy &song = _track_list.get_track(_current_track_index)._songs;

    DataCellTimestamp now(time(NULL));
    song[SFI_loaded_time] = now;
    song[SFI_filename] = DataCell(_filename);
    song[SFI_file_type] = DataCell(_load_type);
    song[SFI_volume_id] = DataCell(_volume.get_volume_id());
    song.record_db();

    music_db->sql_command
      ("INSERT INTO songs_convert (song_id, new_file_type,"
       " new_volume_id, request_time)"
       " VALUES (" +
       DataCell(song.get_primary_key()) + ", " +
       _data[TI_configure][CFI_cdrom_want_type] + ", " +
       _data[TI_configure][CFI_cdrom_want_volume] + ", " +
       now + ")");

    if (!advance_track()) {
      // If that was the last track, close the window.
      delete_self();
    }
  } else {
    if (!_canceled) {
      new InfoMessage("Failure reading CD.");
    }
    delete_self();
  }
}

void CDRecordDialog::
cancel_button_clicked() {
  _canceled = true;
  _cd_reader->stop_reading();
}


void CDRecordDialog::
layout_window() {
  set_default_size(200, 0);
  set_title("Reading CD");

  // The box for the main set of widgets.
  vbox_.pack_start(_box1);
  _box1.show();
  _box1.set_border_width(8);

  _box1.pack_start(_cd_title, true, false, 0);
  _cd_title.show();
  _box1.pack_start(_progress_bar, true, false, 0);
  _progress_bar.show();
  _box1.pack_start(_track_name, true, false, 0);
  _track_name.show();

  // The Cancel button.
  _cancel_button.set_usize(80, 36);
  action_area_.pack_start(_cancel_button, true, false, 0);
  _cancel_button.set_flags(GTK_CAN_DEFAULT);
  _cancel_button.show();

  connect_to_method(_cancel_button.clicked, this, &cancel_button_clicked);
};


bool CDRecordDialog::
advance_track() {
  TableDataProxy &done_song = 
    _track_list.get_track(_current_track_index)._songs;
  _current_length += done_song[SFI_length].get_float();

  // Look for the next track to load.
  _current_track_index++;
  while (_current_track_index < _track_list.get_num_rows() &&
	 !_track_list.get_track(_current_track_index)._load) {
    _current_track_index++;
  }
    
  if (_current_track_index >= _track_list.get_num_rows()) {
    // All done.
    _current_length = _total_length;
    _progress_bar.set_value(_current_length);
    _track_name.set_text(string());
    return false;
  }

  TableDataProxy &next_song = 
    _track_list.get_track(_current_track_index)._songs;

  _progress_bar.set_value(_current_length);
  _track_name.set_text(get_formatted_song_name(next_song));

  return start_reading();
}

bool CDRecordDialog::
start_reading() {
  TableDataProxy &song = _track_list.get_track(_current_track_index)._songs;

  _basename = 
    "cd" + 
    _data[TI_cds][CFI_cddb].get_string() + "_" +
    format_int(_current_track_index + 1) + "_" + 
    format_int(song.get_primary_key()) +
    _extension;

  _filename = _volume.prepare_write_file(_basename);
  if (_filename.empty()) {
    new InfoMessage("Cannot store directly into volume " +
		    _volume.get_description() + ".");
    return false;
  }

  _cd_reader->start_reading(_current_track_index + 1, _filename);
  return true;
}

#endif  // HAVE_CDAUDIO
