/* player.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "player.h"
#include "server.h"
#include "playerDef.h"
#include "commandInterface.h"
#include "extract_words.h"
#include "format.h"

#include <unistd.h>
#include <signal.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>

Player::
Player(int user_socket, Server *server) : 
  _user_socket(user_socket),
  _server(server)
{
  _interface = NULL;
  _verbose = false;
  feedback_user("sfserver");
}

Player::
~Player() {
  stop();
  close(_user_socket);
}

void Player::
read_user_command() {
  char buffer[4096];
  int bytes_read = read(_user_socket, buffer, 4096);
  if (bytes_read > 0) {
    line_up_user_input(string(buffer, bytes_read));
  } else {
    // We read no data; the socket must have closed.
    _server->close_connection(_user_socket);
  }
}

void Player::
interrupt() {
  if (_interface != NULL) {
    delete _interface;
    _interface = NULL;
    feedback_user("interrupt");
  }
}

void Player::
set_filename(const string &filename, int type, double length) {
  if (is_playing()) {
    stop();
  }
  _filename = filename;
  _track_num = 0;
  _def = NULL;
  _is_valid = false;
  _total_time = length;
  _total_frames = 0;

  string extension;
  if (type < 0) {
    _def = PlayerDef::find_def(filename);
  } else {
    _def = PlayerDef::find_def(type);
  }

  if (_def != NULL) {
    init_interface();
  } else {
    if (type < 0) {
      feedback_user("ERROR unknown extension " + extension);
    } else {
      feedback_user("ERROR invalid type " + format_int(type));
    }
    feedback_user("invalid");
  }
}

void Player::
set_cd_track(const string &cdrom_device, int track_num, double length) {
  if (is_playing()) {
    stop();
  }
  _filename = cdrom_device;
  _track_num = track_num;
  _def = NULL;
  _is_valid = false;
  _total_time = length;
  _total_frames = 0;

  _def = PlayerDef::cdrom_def();

  if (_def != NULL) {
    init_interface();
  } else {
    feedback_user("invalid");
  }
}


void Player::
play(double start, double end) {
  _server->set_active_player(this);

  if (is_playing()) {
    stop();
    sleep(1);
  }

  _start_time = start;
  _end_time = end;

  _interface = new CommandInterface
    (_def->get_play_command(_server->get_sound_device(),
			    _filename, _track_num,
			    _start_time, _end_time,
			    _total_frames, _total_time),
     _def->get_play_status(), this);

  _attempt_count = 0;
  _last_reported_time = -1.0;
  feedback_user("play");
}

void Player::
stop() {
  if (_interface != NULL) {
    delete _interface;
    _interface = NULL;
    feedback_user("stop");
  }
}

bool Player::
get_verbose() const {
  return _verbose;
}

bool Player::
is_playing() const {
  return (_interface != NULL);
}

double Player::
get_current_time() const {
  if (_interface == NULL) {
    return 0.0;
  } else {
    return _interface->get_current_time();
  }
}


int Player::
get_total_frames() const {
  return _total_frames;
}

double Player::
get_total_time() const {
  return _total_time;
}

int Player::
get_fd() const {
  if (_interface != NULL) {
    return _interface->get_fd();
  } else {
    return -1;
  }
}

void Player::
read_status() {
  if (_interface != NULL) {
    if (!_interface->read_status()) {
      // The player stopped.  Clear it.

      bool anything = _interface->anything_played();
      delete _interface;
      _interface = NULL;

      if (!anything && _attempt_count == 0) {
	// Hmm, it didn't play anything.  Maybe it wasn't able to get
	// the sound card.  Try again.
	_attempt_count++;
	sleep(1);
	_interface = new CommandInterface
	  (_def->get_play_command(_server->get_sound_device(),
				  _filename, _track_num,
				  _start_time, _end_time,
				  _total_frames, _total_time),
	   _def->get_play_status(), this);
	return;
      }

      if (!anything) {
	// If we couldn't get it to play for whatever reason, it's a
	// failure.  Send "interrupt" as the only way we have to
	// indicate failure.
	feedback_user("interrupt");
      } else {
	feedback_user("end");
      }

    } else if (_interface->get_total_frames() != 0) {
      _total_frames = _interface->get_total_frames();
    }

    if (_def->has_feature(PlayerDef::F_report) && 
	_total_time > 0.0 && 
	_interface != NULL && 
	_interface->anything_played()) {
      double current_time = get_current_time();
      if (_last_reported_time != current_time) {
	feedback_user("time " + format_float(current_time));
	_last_reported_time = current_time;
      }
    }
  }
}

void Player::
feedback_user(string feedback) {
  feedback += "\n";
  write(_user_socket, feedback.data(), feedback.size());
}

void Player::
init_interface() {
  string verify_command = _def->get_verify_command(_filename, _track_num);
  if (!verify_command.empty()) {
    // Fire off a quick command to verify the filename, and possibly
    // count the total number of frames or something.
    CommandInterface fc(verify_command, _def->get_verify_status(), this);
    while (fc.read_status()) {
    }

    _is_valid = fc.anything_played();
    _total_frames = fc.get_total_frames();
    _total_time = fc.get_total_time();

  } else {
    // There's no verify command, so we'll have to assume it's valid.
    _is_valid = true;
  }

  if (_is_valid) {
    string feedback = "ready";

    // Report back the feature set available to the player as well.

    if (_def->has_feature(PlayerDef::F_rstart) && _total_time > 0.0) {
      feedback += " rstart";
    }
    if (_def->has_feature(PlayerDef::F_rstop) && _total_time > 0.0) {
      feedback += " rstop";
    }
    if (_def->has_feature(PlayerDef::F_report) && _total_time > 0.0) {
      feedback += " report";
    }

    feedback_user(feedback);
  } else {
    feedback_user("invalid");
  }
}

void Player::
line_up_user_input(const string &input) {
  _user_input += input;

  int sol = 0;
  int eol = _user_input.find_first_of("\n\r", sol);
  while (eol != string::npos) {
    read_user_line(_user_input.substr(sol, eol - sol));
    sol = eol + 1;
    eol = _user_input.find_first_of("\n\r", sol);
  }

  _user_input = _user_input.substr(sol);
}

void Player::
read_user_line(const string &line) {
  vector<string> words = extract_words(line);

  if (!words.empty()) {
    if (words[0] == "file" && words.size() == 2) {
      set_filename(words[1]);

    } else if (words[0] == "file" && words.size() == 3) {
      set_filename(words[1], atoi(words[2].c_str()));

    } else if (words[0] == "file" && words.size() == 4) {
      set_filename(words[1], atoi(words[2].c_str()), atof(words[3].c_str()));

    } else if (words[0] == "cdrom" && words.size() == 3) {
      set_cd_track(words[1], atoi(words[2].c_str()));

    } else if (words[0] == "cdrom" && words.size() == 4) {
      set_cd_track(words[1], atoi(words[2].c_str()),
		   atof(words[3].c_str()));

    } else if (words[0] == "play" && words.size() == 1) {
      if (!_is_valid) {
	feedback_user("ERROR no valid filename given.");
      } else {
	play(0.0, 1000000.0);
      }

    } else if (words[0] == "play" && words.size() == 2) {
      if (!_is_valid) {
	feedback_user("ERROR no valid filename given.");
      } else {
	play(atof(words[1].c_str()), 1000000.0);
      }

    } else if (words[0] == "play" && words.size() == 3) {
      if (!_is_valid) {
	feedback_user("ERROR no valid filename given.");
      } else {
	play(atof(words[1].c_str()), atof(words[2].c_str()));
      }

    } else if (words[0] == "stop" && words.size() == 1) {
      stop();

    } else if (words[0] == "quit" && words.size() == 1) {
      feedback_user("goodbye");
      _server->close_connection(_user_socket);

    } else if (words[0] == "verbose" && words.size() == 1) {
      _verbose = true;
      feedback_user("verbose 1");

    } else if (words[0] == "verbose" && words.size() == 2) {
      _verbose = (atoi(words[1].c_str()) != 0);
      feedback_user("verbose " + format_int(_verbose));

    } else if (words[0] == "config" && words.size() == 1) {
      PlayerDef::read_config();
      feedback_user("config");

    } else {
      feedback_user("ERROR invalid syntax: " + line);
    }
  }
}
