/* tableDef.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "tableDef.h"
#include "database.h"

#include <map>

TableDef::TableDefs TableDef::_table_defs;
TableDef::FieldsByName TableDef::_all_fields_by_name;

TableDef::
TableDef() {
}

TableDef:: 
~TableDef() {
}

const TableDef *TableDef::
get_table(TableID table_id) {
  if (!is_extension_table(table_id)) {
    return get_system_table(table_id);
  }

  TableDefs::const_iterator edi;
  edi = _table_defs.find(table_id);
  if (edi != _table_defs.end()) {
    return (*edi).second;
  } else {
    TableDef *def = new TableDef;
    def->look_up(table_id);
    _table_defs[table_id] = def;
    return def;
  }
}

const TableDef *TableDef::
reread_table(TableID table_id) {
  if (!is_extension_table(table_id)) {
    return get_system_table(table_id);
  }

  TableDefs::iterator edi;
  edi = _table_defs.find(table_id);
  if (edi != _table_defs.end()) {
    delete (*edi).second;
    TableDef *def = new TableDef;
    def->look_up(table_id);
    (*edi).second = def;
    return def;
  } else {
    TableDef *def = new TableDef;
    def->look_up(table_id);
    _table_defs[table_id] = def;
    return def;
  }
}

void TableDef::
record_db() {
  assert(is_extension_table(_table_id));
  Database *db = get_table_db(_table_id);

  // First, remove any old fields that might or might not be there.
  db->sql_command
    ("DELETE FROM extension_fields WHERE table_id=" + DataCell(_table_id));

  // Now insert a new set of fields..
  Fields::iterator fi;
  for (fi = _fields.begin(); fi != _fields.end(); ++fi) {
    (*fi).second.insert_db(_table_id);
  }
}

bool TableDef::
look_up(TableID table_id) {
  assert(is_extension_table(table_id));

  _table_id = table_id;

  QueryResult qr = music_db->sql_command
    ("SELECT field_id, field_type, foreign_table, default_value, field_name, "
     "label FROM extension_fields "
     "WHERE table_id=" + DataCell(table_id));

  int num_rows = qr.get_num_rows();
  for (int r = 0; r < num_rows; r++) {
    TableField f;
    f._field_id = qr.get_cell(r, 0).get_int();
    f._field_type = (DataCell::Type)qr.get_cell(r, 1).get_int();
    f._foreign_table = (TableID)qr.get_cell(r, 2).get_int(TI_no_table);
    f._default_value = 
      DataCell(qr.get_cell(r, 3).get_string().c_str(), f._field_type);
    f._field_name = qr.get_cell(r, 4).get_string();
    f._label = qr.get_cell(r, 5).get_string();
    _fields[f._field_id] = f;
    _all_fields_by_name[f._field_name] = f;
  }

  if (!_fields.empty() &&
      !music_db->has_table(get_table_name(table_id))) {
    // If we don't have the extension table, create it.
    // Build up a CREATE TABLE command.
    string field_descs;
    Fields::const_iterator fi;
    for (fi = _fields.begin(); fi != _fields.end(); ++fi) {
      const TableField &field = (*fi).second;
      field_descs += ", " + 
	field._field_name + " " + field.get_db_type();
    }

    string create =
      "CREATE TABLE " + get_table_name(table_id) + " (" +
      get_table_key(table_id) + " INT PRIMARY KEY" +
      field_descs + ")";
    music_db->sql_command(create);
  }

  return true;
}

// Given that field_id refers to a valid field in the table
// definition, this returns its reference.
const TableField &TableDef::
get_field(int field_id) const {
  Fields::const_iterator fi;
  fi = _fields.find(field_id);
  assert(fi != _fields.end());
  return (*fi).second;
}

// This searches all the tables in the universe for the field of the
// given name.  Since some field names may be duplicated across
// multiple tables, the particular field that may be defined is
// undefined in this case, so it's best if all duplicated field names
// have exactly the same properties in all tables.

// It may return NULL if there is no such field.

const TableField *TableDef::
find_any_field(const string &name) {
  FieldsByName::const_iterator fi;
  fi = _all_fields_by_name.find(name);
  if (fi == _all_fields_by_name.end()) {
    return NULL;
  } else {
    return &(*fi).second;
  }
}

// Actually, there's presently no real reason to define the primary
// key of a table, since it's already returned by get_table_key().
// But maybe this will change soon; I never thought too much of that
// scheme.

// In the meantime, all this does is add the primary key definition to
// the set of all fields by name.
void TableDef::
define_pkey(const string &field_name, const string &label) {
  TableField t(0, DataCell::DC_foreign_key, _table_id, DataCell(), 
	       field_name, label);
  _all_fields_by_name[field_name] = t;
}

void TableDef::
define_field(int field_id, DataCell::Type field_type, 
	     TableID foreign_table, const DataCell &default_value,
	     const string &field_name, const string &label) {
  TableField t(field_id, field_type, foreign_table, default_value, 
	       field_name, label);

  _fields[field_id] = t;
  _all_fields_by_name[field_name] = t;
}
